/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.api;

// for file processing
import java.io.File;

// for list processing
import java.util.ArrayList;
import java.util.List;

// XML processing
import javax.xml.xpath.XPathConstants;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Abstraction of a <tt>rom.xml</tt> file.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: ROM.java 776 2011-03-18 22:35:07Z yegor256@yahoo.com $
 */
public final class ROM {

    /**
     * DOM Document of the ROM file.
     */
    private final Document document;

    /**
     * ROM utils.
     */
    private final ROMUtils utils = new ROMUtils();

    /**
     * Public ctor, from a file.
     * @param file Absolute <tt>pom.xml</tt> file location
     * @throws ROMException If some ROM parsing problem occurs
     */
    public ROM(final File file) throws ROMException {
        this.document = this.utils.parse(file);
    }

    /**
     * Public ctor, from a file.
     * @param rom Content of <tt>pom.xml</tt>
     * @throws ROMException If some ROM parsing problem occurs
     */
    public ROM(final String rom) throws ROMException {
        this.document = this.utils.parse(rom);
    }

    /**
     * This ROM has class declared?
     * @return Has it?
     * @throws ROMException If some ROM parsing problem occurs
     */
    public boolean hasWorker() throws ROMException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#hasWorker()");
        }
        return this.utils.xpath(
            "/*/class",
            this.document,
            XPathConstants.NODE
        ) != null;
    }

    /**
     * Get instance of the worker.
     * @return The worker
     * @throws ROMException If some ROM parsing problem occurs
     */
    public Object getWorker() throws ROMException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#getWorker()");
        }
        final Node node = (Node) this.utils.xpath(
            "/*/class/text()",
            this.document,
            XPathConstants.NODE
        );
        try {
            return Rempl.getFactory()
                .getLoader()
                .forName(node.getNodeValue())
                .newInstance();
        } catch (LoadingException ex) {
            throw new ROMException(ex);
        } catch (InstantiationException ex) {
            throw new ROMException(ex);
        } catch (IllegalAccessException ex) {
            throw new ROMException(ex);
        }
    }

    /**
     * Get configuration element.
     * @return The configuration
     */
    public ROMConfiguration configuration() {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#configuration()");
        }
        final Node node = (Node) this.utils.xpath(
            "/*/configuration",
            this.document,
            XPathConstants.NODE
        );
        return new ROMConfiguration(node);
    }

    /**
     * Get list of dependencies.
     * @return The artifacts
     * @throws ROMException If some ROM parsing problem occurs
     */
    public List<Artifact> dependencies() throws ROMException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#dependencies()");
        }
        return this.artifacts("/rom/dependencies/dependency");
    }

    /**
     * Get list of reporters.
     * @return The artifacts
     * @throws ROMException If some ROM parsing problem occurs
     */
    public List<Artifact> reporters() throws ROMException {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#reporters()");
        }
        return this.artifacts("/rom/reporters/reporter");
    }

    /**
     * Get list of dependencies.
     * @param root The name of XML root element
     * @return The artifacts
     * @throws ROMException If some ROM parsing problem occurs
     */
    private List<Artifact> artifacts(final String root) throws ROMException {
        final List<Artifact> artifacts = new ArrayList<Artifact>();
        final NodeList nodes = (NodeList) this.utils.xpath(
            root,
            this.document,
            XPathConstants.NODESET
        );
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#artifact('%s'): found %d nodes",
                root,
                nodes.getLength()
            );
        }
        for (int idx = 0; idx < nodes.getLength(); idx += 1) {
            final Node node = nodes.item(idx);
            artifacts.add(
                new Artifact(
                    this.utils.text("groupId/text()", node),
                    this.utils.text("artifactId/text()", node),
                    this.utils.text("version/text()", node)
                )
            );
        }
        return artifacts;
    }

}
