/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.api;

// JAXP
import javax.xml.parsers.DocumentBuilderFactory;

// XPath API
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;

// string processing
import org.apache.commons.lang.StringUtils;

// DOM
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Parser of incoming XML file (in ROM schema).
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: ROMConfiguration.java 692 2011-03-16 18:07:50Z yegor256@yahoo.com $
 */
public final class ROMConfiguration {

    /**
     * Document builder factory, DOM.
     */
    private static DocumentBuilderFactory dbFactory =
        DocumentBuilderFactory.newInstance();

    /**
     * XPath factory, DOM.
     */
    private static XPathFactory xpathFactory =
        XPathFactory.newInstance();

    /**
     * DOM Document beneath it.
     */
    private Document document;

    /**
     * Create configuration.
     * @param node The node
     */
    public ROMConfiguration(final Node node) {
        if (Logger.isTraceEnabled(this)) {
            if (node == null) {
                Logger.trace(this, "#ROMConfiguration(null)");
            } else {
                Logger.trace(this, "#ROMConfiguration(node)");
            }
        }
        try {
            this.document = this.dbFactory.newDocumentBuilder().newDocument();
        } catch (javax.xml.parsers.ParserConfigurationException ex) {
            throw RuntimeProblem.make(ex);
        }
        Node root;
        if (node != null) {
            root = this.document.importNode(node, true);
        } else {
            root = this.document.createElement("configuration");
        }
        this.document.appendChild(root);
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#ROMConfiguration(node): '%s'",
                root.getNodeName()
            );
        }
    }

    /**
     * Append new configuration to the existing one.
     * @param config New configuration
     */
    public void append(final ROMConfiguration config) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#append(config)");
        }
        final Integer replaced = this.migrate(
            // migrate to
            this.document.getDocumentElement(),
            // migrate from
            config.document.getDocumentElement()
        );
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#append(config): replaced %d node(s)",
                replaced
            );
        }
    }

    /**
     * Get text node by XPath.
     * @param addr The address of the node to get
     * @return The string found
     */
    public String get(final String addr) {
        return (String) this.get(addr, String.class);
    }

    /**
     * Get node by XPath.
     * @param addr The address of the node to get
     * @param type Class to convert the object to, through JAXB
     * @return The object found
     */
    public Object get(final String addr, final Class<?> type) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#get(addr, type)");
        }
        final XPath xpath = this.xpathFactory.newXPath();
        NodeList list;
        try {
            list = (NodeList) xpath.evaluate(
                addr,
                this.document,
                XPathConstants.NODESET
            );
        } catch (javax.xml.xpath.XPathExpressionException ex) {
            throw RuntimeProblem.make(ex);
        }
        if (list == null) {
            throw RuntimeProblem.make(addr + " not found");
        }
        final Object value = new ROMDexml().convert(list, type);
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#get('%s', %s): returned '%s'",
                addr,
                type.getName(),
                value.getClass().getName()
            );
        }
        return value;
    }

    /**
     * Set node value by XPath.
     * @param xpath The address of the node to get
     * @param node The node to save (if possible convert it with JAXB)
     */
    public void set(final String xpath, final Object node) {
        if (Logger.isTraceEnabled(this)) {
            Logger.trace(this, "#set(xpath, node)");
        }
        if (!(node instanceof String)) {
            throw RuntimeProblem.make(
                "JAXB is not implemented yet in set(), you can't"
                + " set any types except of java.lang.String"
            );
        }
        final String[] sectors = StringUtils.split(xpath, "/");
        Node root = this.document.getDocumentElement();
        for (String sector : sectors) {
            final NodeList siblings = root.getChildNodes();
            boolean found = false;
            for (int idx = 0; idx < siblings.getLength(); idx += 1) {
                if (siblings.item(idx).getNodeName().equals(sector)) {
                    root = siblings.item(idx);
                    found = true;
                    continue;
                }
            }
            if (found) {
                continue;
            }
            final Node newRoot = this.document.createElement(sector);
            root.appendChild(newRoot);
            root = newRoot;
        }
        root.appendChild(this.document.createTextNode((String) node));
        if (Logger.isDebugEnabled(this)) {
            Logger.debug(
                this,
                "#set('%s', %s): set to '%s'",
                xpath,
                node.getClass().getName(),
                (String) node
            );
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return new ROMUtils().transform(this.document);
    }

    /**
     * Migrate element from source to destination.
     * @param dest Destination node to copy to
     * @param src Source node to copy from
     * @return How many nodes were replaced
     * @see #append(ROMConfiguration)
     */
    private Integer migrate(final Node dest, final Node src) {
        Integer replaced = 0;
        final NodeList srcSiblings = src.getChildNodes();
        final NodeList destSiblings = dest.getChildNodes();
        final StringBuffer textToCopy = new StringBuffer();
        for (int isrc = 0; isrc < srcSiblings.getLength(); isrc += 1) {
            final String srcName = srcSiblings.item(isrc).getNodeName();
            // if it's a text node - just remember it for the future
            if ("#text".equals(srcName)) {
                textToCopy.append(srcSiblings.item(isrc).getNodeValue());
                continue;
            }
            Node migrateTo = null;
            for (Integer idst = 0; idst < destSiblings.getLength(); idst += 1) {
                if (destSiblings.item(idst).getNodeName().equals(srcName)) {
                    migrateTo = destSiblings.item(idst);
                }
            }
            if (migrateTo == null) {
                replaced += 1;
                migrateTo = dest.appendChild(
                    this.document.createElement(srcName)
                );
            }
            replaced += this.migrate(migrateTo, srcSiblings.item(isrc));
        }
        // if we didn't copy any XML elements, and have some
        // text to copy - do it now
        if (replaced == 0 && textToCopy.length() > 0) {
            replaced += this.copyText(dest, textToCopy.toString());
        }
        return replaced;
    }

    /**
     * Copy text to destination.
     * @param dest Destination node to copy to
     * @param txt Text to copy
     * @return How many nodes were replaced
     * @see #migrate(Node, Node)
     */
    private Integer copyText(final Node dest, final String txt) {
        // delete all nodes at destination
        final NodeList destSiblings = dest.getChildNodes();
        for (Integer idx = 0;
            idx < destSiblings.getLength(); idx += 1) {
            dest.removeChild(destSiblings.item(idx));
        }
        // copy just one text node
        dest.appendChild(this.document.createTextNode(txt));
        return 1;
    }

}
