/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.api;

// for URL manipulations
import java.net.URL;

// for manipulations with files
import org.apache.commons.io.IOUtils;

/**
 * Entry point of the package, instantiator of
 * a {@link RemplFactory}.
 *
 * <p>In order to find the name of the
 * class with implemenation we are using <tt>META-INF/services</tt>
 * directory. We're trying to find there a file named
 * <tt>"com.rempl.api.RemplFactory"</tt> and read its content. It
 * has to include the name of the class. If the file is absent a
 * {@link RuntimeException} will be thrown.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: Rempl.java 573 2011-03-13 10:13:17Z yegor256@yahoo.com $
 * @see <a href="http://download.oracle.com/javase/1.4.2/docs/guide/jar/jar.html#Service%20Provider">JAR Service Provider</a>
 */
public final class Rempl {

    /**
     * This is where the name of the implementation class
     * has to be stored.
     */
    private static final String RESOURCE =
        "/META-INF/services/com.rempl.api.RemplFactory";

    /**
     * The implementation.
     */
    private static RemplFactory factory;

    /**
     * Private ctor, to avoid instantiation.
     */
    private Rempl() {
        // intentionally empty
    }

    /**
     * Returns an instance of {@link RemplFactory}.
     * @return The instance of {@link RemplFactory}
     */
    public static RemplFactory getFactory() {
        if (Rempl.factory == null) {
            Rempl.factory = Rempl.instantiate();
        }
        return Rempl.factory;
    }

    /**
     * Clean internal state. The method is only for usage
     * inside this package and mostly for unit testing.
     */
    protected static void clean() {
        Rempl.factory = null;
    }

    /**
     * Creates and returns an instance of {@link RemplFactory}.
     * @return The instance of {@link RemplFactory}
     * @see #getFactory()
     */
    private static RemplFactory instantiate() {
        final URL svc = Rempl.class.getResource(Rempl.RESOURCE);
        if (svc == null) {
            throw RuntimeProblem.make(
                "Name of implementation class has to be specified "
                + "inside file %s (which is not available now in classpath)",
                Rempl.RESOURCE
            );
        }
        if (Logger.isDebugEnabled(Rempl.class)) {
            Logger.debug(Rempl.class, "#instantiate(): URL = %s", svc);
        }
        Object object;
        try {
            object = Rempl.makeClass(svc).newInstance();
        } catch (java.lang.InstantiationException ex) {
            throw RuntimeProblem.make(ex);
        } catch (java.lang.IllegalAccessException ex) {
            throw RuntimeProblem.make(ex);
        }
        if (!(object instanceof RemplFactory)) {
            throw RuntimeProblem.make(
                "%s doesn't inherit from RemplFactory",
                object.getClass().getName()
            );
        }
        return (RemplFactory) object;
    }

    /**
     * Creates a Class.
     * @param url The URL where the file is
     * @return The class that has to be instantiated
     * @see #instantiate()
     */
    private static Class makeClass(final URL url) {
        String name;
        try {
            name = IOUtils.toString(url.openStream());
        } catch (java.io.IOException ex) {
            throw RuntimeProblem.make(ex);
        }
        if (Logger.isDebugEnabled(Rempl.class)) {
            Logger.debug(Rempl.class, "#instantiate(): name = %s", name);
        }
        Class impl;
        try {
            impl = Class.forName(name);
        } catch (java.lang.ClassNotFoundException ex) {
            throw RuntimeProblem.make(ex);
        }
        if (Logger.isDebugEnabled(Rempl.class)) {
            Logger.debug(
                Rempl.class,
                "#instantiate(): Class = %s",
                impl.getName()
            );
        }
        return impl;
    }

}
